/* Use x-Kernel versions of the protocols */
#include "xkernel.h"
#include "xpf.h"
#include "xpf_internal.h"
#include "xchip.h"
#include "pattern.h"

#if	PROTOTYPE
/* Internal functions */
int	setup_xip_program(XObj, int, unsigned int);
int	setup_xip_filter(XObj, int, unsigned int);
int	remove_xip_program(XObj, int, unsigned int);
int	remove_xip_filter(XObj, int, unsigned int);
int	setup_udpip_xchip(XObj, int, unsigned int);
int	remove_udpip_xchip(XObj, int, unsigned int);

/* User callable */
int	setup_udpip_pattern(XObj, int, unsigned int);
int	remove_udpip_pattern(XObj, int, unsigned int);
int	setup_tcpip_pattern(XObj, int, unsigned int);
int	remove_tcpip_pattern(XObj, int, unsigned int);
#endif	PROTOTYPE

#define CHECK_MOREFRAGS	1
#if	CHECK_MOREFRAGS
/* ipaddr is ip-destaddr; protid is protocol above ip; morefragval is IP_MOREFRAG_MASK if programs are there, 0 if not */
#define create_ip_filter(fip, ipaddr, protid, morefragval)\
{\
	xpfCreateFilter((fip));\
	xpfInitializeSelector(&(fip)->extractor.selector, IP_DESTADDR_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(fip)->extractor, XPF_DATA_INT, IP_DESTADDR_MASK);\
	xpfInitializeFilter((fip), (ipaddr), NULL);\
\
	xpfCreateFilter((fip)->next);\
	xpfInitializeSelector(&(fip)->next->extractor.selector, IP_PROTID_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(fip)->next->extractor, XPF_DATA_CHAR, 0xff);\
	xpfInitializeFilter((fip)->next, (protid), NULL);\
\
	xpfCreateFilter((fip)->next->next);\
	xpfInitializeSelector(&(fip)->next->next->extractor.selector, IP_MOREFRAG_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(fip)->next->next->extractor, XPF_DATA_SHORT, htons(IP_MOREFRAG_MASK));\
	xpfInitializeFilter((fip)->next->next, htons(morefragval), NULL);\
\
	/* if (morefragval) { */\
	/* } */\
}
#else	CHECK_MOREFRAGS
/* The parameter morefragval is not used at all. The same stuff gets set up for both programs and filters */
#define create_ip_filter(fip, ipaddr, protid, morefragval)\
{\
	xpfCreateFilter((fip));\
	xpfInitializeSelector(&(fip)->extractor.selector, IP_DESTADDR_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(fip)->extractor, XPF_DATA_INT, IP_DESTADDR_MASK);\
	xpfInitializeFilter((fip), (ipaddr), NULL);\
\
	xpfCreateFilter((fip)->next);\
	xpfInitializeSelector(&(fip)->next->extractor.selector, IP_PROTID_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(fip)->next->extractor, XPF_DATA_CHAR, 0xff);\
	xpfInitializeFilter((fip)->next, (protid), NULL);\
}
#endif	CHECK_MOREFRAGS

#define create_ip_program(oip)\
{\
	xpfCreateProgram((oip));\
	xpfInitializeSelector(&(oip)->extractor.selector, IP_DESTADDR_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(oip)->extractor, XPF_DATA_INT, IP_DESTADDR_MASK);\
	xpfInitializeProgram((oip), NULL);\
	xpfCreateProgram((oip)->next);\
	xpfInitializeSelector(&(oip)->next->extractor.selector, IP_MSGID_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(oip)->next->extractor, XPF_DATA_SHORT, SHORT_MASK);\
	xpfInitializeProgram((oip)->next, NULL);\
}

/* pnext is the pattern for the protocol above ip */
#define create_ip_pattern(pip, ipaddr, protid, morefragval, pnext)\
{\
	XPFFilter	*fip;\
	XPFProgram	*oip = NULL;\
\
	xpfCreatePattern((pip));\
	create_ip_filter(fip, (ipaddr), (protid), (morefragval));\
	if ((morefragval) != 0) create_ip_program(oip);\
	xpfInitializeFixedHeader(&(pip)->header, IP_HEADER_SIZE);\
	xpfInitializePattern((pip), (fip), (oip), FLAGS, ALARM, (pnext));\
}

/* udp and tcp patterns are terminal, so no next-protocol-id and next-pattern-pointer */

#define create_udp_filter(fudp, udpport) \
{\
	xpfCreateFilter(fudp);\
	xpfInitializeSelector(&(fudp)->extractor.selector, UDP_DESTPORT_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(fudp)->extractor, XPF_DATA_SHORT, SHORT_MASK);\
	xpfInitializeFilter((fudp), (udpport), NULL);\
}

#define create_udp_pattern(pudp, udpport) \
{\
	XPFFilter	*fudp;\
\
	xpfCreatePattern(pudp);\
	create_udp_filter((fudp), (udpport));\
	xpfInitializeFixedHeader(&(pudp)->header, UDP_HEADER_SIZE);\
	xpfInitializePattern((pudp), (fudp), NULL, FLAGS, ALARM, NULL);\
}

#define create_tcp_filter(ftcp, tcpport) \
{\
	xpfCreateFilter(ftcp);\
	xpfInitializeSelector(&(ftcp)->extractor.selector, TCP_DESTPORT_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(ftcp)->extractor, XPF_DATA_SHORT, SHORT_MASK);\
	xpfInitializeFilter((ftcp), (tcpport), NULL);\
}

#define create_tcp_pattern(ptcp, tcpport) \
{\
	XPFFilter	*ftcp;\
\
	xpfCreatePattern(ptcp);\
	create_tcp_filter((ftcp), (tcpport));\
	xpfInitializeFixedHeader(&(ptcp)->header, TCP_HEADER_SIZE);\
	xpfInitializePattern((ptcp), (ftcp), NULL, FLAGS, ALARM, NULL);\
}

/* This is same as TCP filter but has one extra filter to do the protocol level dispatch missing above since that was terminal */
#define create_dummy_filter(fdummy, dummyport) \
{\
	xpfCreateFilter(fdummy);\
	xpfInitializeSelector(&(fdummy)->extractor.selector, TCP_DESTPORT_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(fdummy)->extractor, XPF_DATA_SHORT, SHORT_MASK);\
	xpfInitializeFilter((fdummy), (dummyport), NULL);\
	xpfCreateFilter((fdummy)->next);\
	xpfInitializeSelector(&(fdummy)->next->extractor.selector, TCP_NEXTPROTL_OFFSET, 0 /*, 1 if countunit */);\
	xpfInitializeExtractor(&(fdummy)->next->extractor, XPF_DATA_CHAR, 0xff);\
	xpfInitializeFilter((fdummy)->next, TCP_NEXTPROTL_ID, NULL);\
}

#define create_dummy_pattern(pdummy, dummyport) \
{\
	XPFFilter	*fdummy;\
\
	xpfCreatePattern(pdummy);\
	create_dummy_filter((fdummy), (dummyport));\
	xpfInitializeFixedHeader(&(pdummy)->header, TCP_HEADER_SIZE);\
	xpfInitializePattern((pdummy), (fdummy), NULL, FLAGS, ALARM, NULL);\
}

int
setup_variable_filter(s, xport, ipaddr, xnum, info)
	XObj		s;
	int		xport;
	unsigned int	ipaddr;
	int		xnum;
	XPFInformation	info;
{
	XPFPattern	*pip, *px, *pcur=NULL;
	XPFSetupArg	a;
	int		ret;

	for (; xnum>0; xnum--) {
		create_dummy_pattern(px, xport);
		px->next = pcur;
		pcur = px;
	}
	create_ip_pattern(pip, ipaddr, TCP_ONIP_PROTID, 0, pcur);

#if	0
	printf("setup_variable_filter's pattern-list:\n");
	xpfPrintPatternList(pip);
	printf("\n");
#endif	0

	a.pattern = pip;
	a.information = info;
	ret = xControl(s, XPF_SETUP, (char *)&a, sizeof(char *));
	xpfDestroyPatternList(pip);
	return ret;
}

int
create_variable_filter(xport, ipaddr, xnum, pat)
	int		xport;
	unsigned int	ipaddr;
	int		xnum;
	XPFPattern	**pat;
{
	XPFPattern	*pip, *px, *pcur = NULL;

	*pat = NULL;
	for (; xnum>0; xnum--) {
		create_dummy_pattern(px, xport);
		px->next = pcur;
		pcur = px;
	}
	create_ip_pattern(pip, ipaddr, TCP_ONIP_PROTID, 0, pcur);
	*pat = pip;

#if	1
	printf("create_variable_filter's pattern-list:\n");
	xpfPrintPatternList((*pat));
	printf("\n");
#endif	1
	return 0;
}

int
setup_xip_program(s, xport, ipaddr, x, info)
	XObj		s;
	int		xport;
	unsigned int	ipaddr;
	char		*x;
	XPFInformation	info;
{
	XPFPattern	*pip, *px;
	XPFSetupArg	a;
	int		ret;

	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) {
		create_udp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, UDP_ONIP_PROTID, IP_MOREFRAG_MASK, px);
	}
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) {
		create_tcp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, TCP_ONIP_PROTID, IP_MOREFRAG_MASK, px);
	}
	else return -1;

#if	0
	printf("setup_xip_program's pattern-list:\n");
	xpfPrintPatternList(pip);
	printf("\n");
#endif	0

	a.pattern = pip;
	a.information = info;
	ret = xControl(s, XPF_SETUP, (char *)&a, sizeof(char *));
	xpfDestroyPatternList(pip);
	return ret;
}

int
setup_xip_filter(s, xport, ipaddr, x, info)
	XObj		s;
	int		xport;
	unsigned int	ipaddr;
	char		*x;
	XPFInformation	info;
{
	XPFPattern	*pip, *px;
	XPFSetupArg	a;
	int		ret;

	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) {
		create_udp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, UDP_ONIP_PROTID, 0, px);
	}
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) {
		create_tcp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, TCP_ONIP_PROTID, 0, px);
	}
	else return -1;

#if	0
	printf("setup_xip_filter's pattern-list:\n");
	xpfPrintPatternList(pip);
	printf("\n");
#endif	0

	a.pattern = pip;
	a.information = info;
	ret = xControl(s, XPF_SETUP, (char *)&a, sizeof(char *));
	xpfDestroyPatternList(pip);
	return ret;
}

int
create_xip_filter(xport, ipaddr, x, pat)
	int		xport;
	unsigned int	ipaddr;
	char		*x;
	XPFPattern	**pat;
{
	XPFPattern	*pip, *px;

	*pat = NULL;
	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) {
		create_udp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, UDP_ONIP_PROTID, 0, px);
	}
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) {
		create_tcp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, TCP_ONIP_PROTID, 0, px);
	}
	else return -1;

	*pat = pip;
#if	1
	printf("create_xip_filter's pattern-list:\n");
	xpfPrintPatternList((*pat));
	printf("\n");
#endif	1
	return 0;
}

int
remove_xip_program(s, xport, ipaddr, x)
	XObj		s;
	int		xport;
	unsigned int	ipaddr;
	char		*x;
{
	XPFPattern	*pip, *px;
	XPFRemoveArg	c;
	int		ret;

	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) {
		create_udp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, UDP_ONIP_PROTID, IP_MOREFRAG_MASK, px);
	}
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) {
		create_tcp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, TCP_ONIP_PROTID, IP_MOREFRAG_MASK, px);
	}
	c.pattern = pip;
	c.information = (XPFInformation) XPF_INFO_TEST;
	xControl(s, XPF_REMOVE, (char *)&c, sizeof(char *));
	xpfDestroyPatternList(pip);

	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) create_udp_pattern(px, xport)
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) create_tcp_pattern(px, xport)
	c.pattern = px;
	c.information = (XPFInformation) XPF_INFO_TEST;
	xControl(s, XPF_REMOVE, (char *)&c, sizeof(char *));
	xpfDestroyPatternList(px);
	return 1;
}

int
remove_xip_filter(s, xport, ipaddr, x)
	XObj		s;
	int		xport;
	unsigned int	ipaddr;
	char		*x;
{
	XPFPattern	*pip, *px;
	XPFRemoveArg	c;
	int		ret;

	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) {
		create_udp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, UDP_ONIP_PROTID, 0, px);
	}
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) {
		create_tcp_pattern(px, xport);
		create_ip_pattern(pip, ipaddr, TCP_ONIP_PROTID, 0, px);
	}
	c.pattern = pip;
	c.information = (XPFInformation) XPF_INFO_TEST;
	xControl(s, XPF_REMOVE, (char *)&c, sizeof(char *));
	xpfDestroyPatternList(pip);

	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) create_udp_pattern(px, xport)
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) create_tcp_pattern(px, xport)
	c.pattern = px;
	c.information = (XPFInformation) XPF_INFO_TEST;
	xControl(s, XPF_REMOVE, (char *)&c, sizeof(char *));
	xpfDestroyPatternList(px);
	return 1;
}

/* This part of it sets up the xchip stuff using two other control ops */

FILTER	chip;

int
setup_xip_xchip(s, xport, ipaddr, x)
	XObj		s;
	int		xport;
	unsigned int	ipaddr;
	char		*x;
{
	bzero((char *)&chip, sizeof(FILTER));
	chip.control = 0xf;
	chip.type = SINGLE;
	/* IP destination address */
	chip.mask[0] = (IP_DESTADDR_MASK & UPPER_SHORT_MASK) >> 16;
	chip.mask[1] = IP_DESTADDR_MASK & SHORT_MASK;
	chip.offset[0] = IP_DESTADDR_OFFSET;
	chip.offset[1] = IP_DESTADDR_OFFSET + sizeof(unsigned short);
	chip.value[0] = (ipaddr & UPPER_SHORT_MASK) >> 16;
	chip.value[1] = ipaddr & SHORT_MASK;
	/* X destination port */
	chip.mask[2] = SHORT_MASK;
	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) chip.offset[2] = IP_HEADER_SIZE + UDP_DESTPORT_OFFSET;
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) chip.offset[2] = IP_HEADER_SIZE + TCP_DESTPORT_OFFSET;
	else return -1;
	chip.value[2] = xport;
	/* IP more fragments */
	chip.mask[3] = IP_MOREFRAG_MASK;
	chip.offset[3] = IP_MOREFRAG_OFFSET;
	chip.value[3] = 0;
	if (-1 == xControl(s, XCHIP_SETUP, (char *)&chip, sizeof(chip))) return -1;

	chip.type = DUAL;
	chip.value[3] = IP_MOREFRAG_MASK;
	if (-1 == xControl(s, XCHIP_SETUP, (char *)&chip, sizeof(chip))) return -1;

	bzero((char *)&chip, sizeof(FILTER));
	chip.control = 0x7;
	chip.type = SINGLE;
	/* IP message-id extraction */
	chip.mask[0] = SHORT_MASK;
	chip.offset[0] = IP_MSGID_OFFSET;
	chip.value[0] = 0;
	/* IP destination address */
	chip.mask[1] = (IP_DESTADDR_MASK & UPPER_SHORT_MASK) >> 16;
	chip.mask[2] = IP_DESTADDR_MASK & SHORT_MASK;
	chip.offset[1] = IP_DESTADDR_OFFSET;
	chip.offset[2] = IP_DESTADDR_OFFSET + sizeof(unsigned short);
	chip.value[1] = (ipaddr & UPPER_SHORT_MASK) >> 16;
	chip.value[2] = ipaddr & SHORT_MASK;
	return xControl(s, XCHIP_SETUP, (char *)&chip, sizeof(chip));
}

int
remove_xip_xchip(s, xport, ipaddr, x)
	XObj		s;
	int		xport;
	unsigned int	ipaddr;
	char		*x;
{
	bzero((char *)&chip, sizeof(FILTER));
	chip.control = 0xf;
	chip.type = SINGLE;
	/* IP destination address */
	chip.mask[0] = (IP_DESTADDR_MASK & UPPER_SHORT_MASK) >> 16;
	chip.mask[1] = IP_DESTADDR_MASK & SHORT_MASK;
	chip.offset[0] = IP_DESTADDR_OFFSET;
	chip.offset[1] = IP_DESTADDR_OFFSET + sizeof(unsigned short);
	chip.value[0] = (ipaddr & UPPER_SHORT_MASK) >> 16;
	chip.value[1] = ipaddr & SHORT_MASK;
	/* X destination port */
	chip.mask[2] = SHORT_MASK;
	if (!strcmp(x, "udp") || !strcmp(x, "UDP")) chip.offset[2] = IP_HEADER_SIZE + UDP_DESTPORT_OFFSET;
	else if (!strcmp(x, "tcp") || !strcmp(x, "TCP")) chip.offset[2] = IP_HEADER_SIZE + TCP_DESTPORT_OFFSET;
	else return -1;
	chip.value[2] = xport;
	/* IP more fragments */
	chip.mask[3] = IP_MOREFRAG_MASK;
	chip.offset[3] = IP_MOREFRAG_OFFSET;
	chip.value[3] = 0;
	if (-1 == xControl(s, XCHIP_REMOVE, (char *)&chip, sizeof(chip))) return -1;

	chip.type = DUAL;
	chip.value[3] = IP_MOREFRAG_MASK;
	if (-1 == xControl(s, XCHIP_REMOVE, (char *)&chip, sizeof(chip))) return -1;

	bzero((char *)&chip, sizeof(FILTER));
	chip.control = 0x7;
	chip.type = SINGLE;
	/* IP message-id extraction */
	chip.mask[0] = SHORT_MASK;
	chip.offset[0] = IP_MSGID_OFFSET;
	chip.value[0] = 0;
	/* IP destination address */
	chip.mask[1] = (IP_DESTADDR_MASK & UPPER_SHORT_MASK) >> 16;
	chip.mask[2] = IP_DESTADDR_MASK & SHORT_MASK;
	chip.offset[1] = IP_DESTADDR_OFFSET;
	chip.offset[2] = IP_DESTADDR_OFFSET + sizeof(unsigned short);
	chip.value[1] = (ipaddr & UPPER_SHORT_MASK) >> 16;
	chip.value[2] = ipaddr & SHORT_MASK;
	return xControl(s, XCHIP_REMOVE, (char *)&chip, sizeof(chip));
}

/*
 * These are the only user interface routines.
 * If cache is not specified, then programs are not generated.
 */

int
setup_variable_pattern(s, varport, varaddr, varnum)
	XObj		s;
	int		varport;
	unsigned int	varaddr;
	int		varnum;
{
	if (setup_variable_filter(s, varport, varaddr, varnum, (XPFInformation) XPF_INFO_TEST) < 0) return -1;
	return 1;
}

int
setup_udpip_pattern(s, udpport, ipaddr, cache)
	XObj		s;
	int		udpport;
	unsigned int	ipaddr;
	int		cache;
{
	if (setup_xip_filter(s, udpport, ipaddr, "udp", (XPFInformation) XPF_INFO_TEST) < 0) return -1;
	if (cache && setup_xip_program(s, udpport, ipaddr, "udp", (XPFInformation) XPF_INFO_TEST) < 0) return -1;
	setup_xip_xchip(s, udpport, ipaddr, "udp");	/* only a FIFO cache */
	return 1;
}

int
remove_udpip_pattern(s, udpport, ipaddr, cache)
	XObj		s;
	int		udpport;
	unsigned int	ipaddr;
{
	if (remove_xip_filter(s, udpport, ipaddr, "udp") < 0) return -1;
	if (cache && remove_xip_program(s, udpport, ipaddr, "udp") < 0) return -1;
	remove_xip_xchip(s, udpport, ipaddr, "udp");	/* only a FIFO cache */
	return 1;
}

int
setup_tcpip_pattern(s, tcpport, ipaddr, cache, pf)
	XObj		s;
	int		tcpport;
	unsigned int	ipaddr;
	int		cache;
	long		pf;
{
	if (setup_xip_filter(s, tcpport, ipaddr, "tcp", (XPFInformation) XPF_INFO_TEST) < 0) return -1;
	if (cache && setup_xip_program(s, tcpport, ipaddr, "tcp", (XPFInformation) XPF_INFO_TEST) < 0) return -1;
	setup_xip_xchip(s, tcpport, ipaddr, "tcp");	/* only a FIFO cache */
	return 1;
}

int
remove_tcpip_pattern(s, tcpport, ipaddr, cache)
	XObj		s;
	int		tcpport;
	unsigned int	ipaddr;
{
	if (remove_xip_filter(s, tcpport, ipaddr, "tcp") < 0) return -1;
	if (cache && remove_xip_program(s, tcpport, ipaddr, "tcp") < 0) return -1;
	remove_xip_xchip(s, tcpport, ipaddr, "tcp");	/* only a FIFO cache */
	return 1;
}

